//-------+---------+---------+---------+---------+---------+---------+--------=
//
// File:	bobutil.c								Author:		Bob Harris
//
// Support code for generic C applications
//
//----------

//----------
//
// bobutil--
//	Miscellaneous utility routines.
//
//----------

#include <stdlib.h>				// standard C stuff
#define  true  1				// oughta be in stdlib.h
#define  false 0				// oughta be in stdlib.h
#include <stdio.h>				// standard C i/o stuff
#include <ctype.h>				// standard C upper/lower stuff
#include <limits.h>				// standard C value limit stuff
#include "util.h"				// Scott Schwartz's handy utility stuff

#define  __bobutil_owner		// make this the owner of this module's globals
#include "bobutil.h"			// interface to this module

//----------
//
// parse_seq_name, parse_seq_name2--
//  Parse a sequence header to extract some useful information.
//
// We expect the sequence header to have a format like this:
//  /depot/.../chr1.nib 26793793 26794393 + BC000995
// The useful information is
//  name:        chr1
//  first index: 26793793
//  last index:  26794393
//	strand:      +
//	comment:     BC000995
//
//----------
//
// Arguments:
//  char*   h:				The sequence header.
//	int		commatize:		true  => include commas in first and last indexes
//							         .. in longName.
//  char**  name:			Place to return the sequence name.  The caller is
//							.. responsible for deallocating this string.
//							.. Unlike the other arguments, this CANNOT be
//							.. NULL.
//  char**  longName:		Place to return the long version of the sequence
//							.. name, which includes the first and last index,
//							.. direction, and comment.  The caller is
//							.. responsible for deallocating this string).
//							.. This can be NULL if the caller has no interest.
//	char*			chr:	Place to return the encoded chromosome (as per
//							.. str_to_chromosome()).  If the header doesn't
//							.. provide this information, the value 0 is used.
//							.. This can be NULL if the caller has no interest.
//  unsigned long*	first:	Place to return the first index.  If the header
//							.. doesn't provide this information, the value 1
//							.. is used.  This can be NULL if the caller has
//							.. no interest.
//  unsigned long*	last:	Place to return the last index.  If the header
//							.. doesn't provide this information, the value 0
//							.. is used.  This can be NULL if the caller has
//							.. no interest.
//	char*			strand:	Place to return the direction/strand
//							..  '+' => forward strand
//							..  '-' => reverse strand
//							..  \0  => strand not provided (null character)
//							.. This can be NULL if the caller has no interest.
//	char**			extra:  Place to return any other info that's attached to
//							.. the header.  This can be NULL if the caller has
//							.. no interest.  The caller is responsible for
//							.. deallocating this string.
//
// Returns:
//  (nothing)
//
//----------

void parse_seq_name
   (char*			h,
	char**			name,
	char**			longName,
	char*			_chr,
	unsigned long*	_first,
	unsigned long*	_last,
	char*			_strand)
	{
	parse_seq_name2 (h, true,
	                 name, longName, _chr, _first, _last, _strand, NULL);
	}

void parse_seq_name2
   (char*			h,
	int				commatize,
	char**			name,
	char**			longName,
	char*			_chr,
	unsigned long*	_first,
	unsigned long*	_last,
	char*			_strand,
	char**			_extra)
	{
	char*			s, *loc, *extra;
	char			c;
	unsigned long	first, last;
	char			strand = 0;
	int				numItems, size, items1Len, items2Len;
	char*			fs, *ls, *strandStr;

	if (h == NULL)
		{
		*name    = copy_string ("unnamed sequence");
		c        = 0;
		numItems = 0;
		extra    = "";
		goto return_stuff;
		}

	// extract the name

	s = h;
	if (*s == '>') s++;
	s = skip_whitespace (s);

	while (*s != 0)
		{
		loc = strpbrk (s, "/ \t");
		if (loc == NULL) break;
		if (*loc != '/') break;
		s = loc+1;
		}

	loc = strpbrk (s, ". \t");
	if (loc == NULL)
		{ *name = copy_string (s);            s += strlen (s);           }
	else
		{ *name = copy_substring (s, loc-s);  s = skip_blackspace (loc); }

	// extract the chromosome

	c = str_to_chromosome (*name);

	// extract the indexes

	s = skip_whitespace (s);

	numItems = sscanf (s, "%lu%n %lu%n",
	                      &first, &items1Len, &last, &items2Len);

	if      (numItems == 1) s += items1Len;
	else if (numItems == 2) s += items2Len;

	// extract the direction/strand

	strand = 0;

	s = skip_whitespace (s);

	if      (*s == '-') { s++;  strand = '-'; }
	else if (*s == '+') { s++;  strand = '+'; }

	// extract the comment (it's the rest of the line)

	extra = skip_whitespace (s);

	if (*extra == 0)
		extra = NULL;

	// return indexes to the caller

return_stuff:

	if (numItems < 1) first = 1;
	if (numItems < 2) last  = 0;

	if (longName != NULL)
		{
		if (commatize)
			{
			fs = copy_string (comma_string (first));
			ls = copy_string (comma_string (last));
			}
		else
			{
			fs = copy_string (number_string (first));
			ls = copy_string (number_string (last));
			}

		size = strlen(*name) + 1				// (name)
		     + strlen(fs) + 1 + strlen(ls)		// (indexes)
		     + 4;								// (direction)
		if (extra != NULL)
			size += 1 + strlen(extra);			// (comment)

		*longName = ckalloc ((size+1)*sizeof(char));

		switch (strand)
			{
			default:
			case '+': strandStr = ">>>";   break;
			case '-': strandStr = "<<<";   break;
			case 0:   strandStr = "<<>>";  break;
			}

		if (extra == NULL)
			sprintf (*longName, "%s:%s-%s %s",
			                    *name, fs, ls, strandStr);
		else
			sprintf (*longName, "%s:%s-%s %s %s",
			                    *name, fs, ls, strandStr, extra);

		ckfree (fs);
		ckfree (ls);
		}

	if (_chr    != NULL) *_chr    = c;
	if (_first  != NULL) *_first  = first;
	if (_last   != NULL) *_last   = last;
	if (_strand != NULL) *_strand = strand;

	if (_extra  != NULL)
		{
		if (extra == NULL) *_extra = NULL;
		              else *_extra = copy_string (extra);
		}

	}

//----------
//
// comma_string, number_string--
//  Create a textual representation of a number, with or without commas.
//
//----------
//
// Arguments:
//  unsigned long v: The number to represent.
//
// Returns:
//  A string containing the representation of the number.  This string is
//  actually static data belonging to this routine, so the caller must copy
//  it if more than one converted number is to be used simultaneously.
//
//----------

char* comma_string
   (unsigned long	v)
	{
	static			char s[31];
	int				digits;
	char*			ss;

	// handle zero as a special case

	if (v == 0)
		{ s[0] = '0';  s[1] = 0;  return s; }

	// build the string in reverse

	ss = s + sizeof(s) - 1;
	*ss = 0;

	digits = 0;
	for ( ; v>9 ; v/=10)
		{
		*(--ss) = '0' + (v % 10);
		if ((++digits % 3) == 0) *(--ss) = ',';
		}

	*(--ss) = '0' + v;

	return ss;
	}

char* number_string
   (unsigned long	v)
	{
	static			char s[31];
	char*			ss;

	// handle zero as a special case

	if (v == 0)
		{ s[0] = '0';  s[1] = 0;  return s; }

	// build the string in reverse

	ss = s + sizeof(s) - 1;
	*ss = 0;

	for ( ; v>9 ; v/=10)
		*(--ss) = '0' + (v % 10);

	*(--ss) = '0' + v;

	return ss;
	}

//----------
//
// skip_whitespace--
//  Skip to the next non-whitespace character in a string.
// skip_blackspace--
//  Skip to the next whitespace character in a string.
//
//----------
//
// Arguments:
//  const char*	s: The string.
//
// Returns:
//	A pointer to the next such character in a string (or to the end of the
//	string.
//
//----------

char* skip_whitespace
   (const char*	s)
	{
	for ( ; (*s!=0)&&(isspace(*s)) ; s++)
		;

	return (char*) s;
	}

char* skip_blackspace
   (const char*	s)
	{
	for ( ; (*s!=0)&&(!isspace(*s)) ; s++)
		;

	return (char*) s;
	}

//----------
//
// string_to_upper--
//  Convert a string to upper case, in place.
//
//----------
//
// Arguments:
//	char*	s:	The string to convert.
//
// Returns:
//	A pointer to the string.  Any lower case characters will have been
//	converted to upper case.  All other characters are unchanged (well,
//	techincally all characters are mapped through the standard c toupper()).
//
//----------

char* string_to_upper
   (char*	s)
	{
	char*	ss;

	for (ss=s ; *ss!=0 ; ss++)
		*ss = toupper (*ss);

	return s;
	}

//----------
//
// string_to_int--
//  Convert a string to an integer.
//
//----------
//
// Arguments:
//	char*	s:		The string to convert.
//  int*	val:	Place to return the value.  This will only be written to
//					.. if the conversion is successsful.
//
// Returns:
//  true if the string contained a valid integer and nothing else;  false
//  otherwise.
//
// Notes:
//  - This routine also recognizes strings like "13K" and "24M", with K and M
//    being abbreviations for thousand and million (decimal, not the binary
//    approximations).
//
//----------

int string_to_int
   (char*	s,
	int*	_val)
	{
	int		val;
	int		items;
	char	units, extra;
	int		mult;

	// if the string is NULL or blank, it's invalid

	if (s == NULL) return false;

	while (isspace (*s)) s++;
	if (*s == 0) return false;

	// parse the number, making sure there's nothing meaningless in the string

	items = sscanf (s, "%d%c%c", &val, &units, &extra);

	if ((items != 1) && (items != 2)) return false;

	// apply a units multiplier, being careful to avoid overflow

	if (items == 2)
		{
		switch (toupper(units))
			{
			case 'K': mult = 1000;       break;
			case 'M': mult = 1000000;    break;
			case 'G': mult = 1000000000; break;
			default: return false;
			}

		if (val >= 0) { if (val  > INT_MAX / mult) return false; }
		         else { if (-val > INT_MAX / mult) return false; }

		val *= mult;
		}

	// success!

	*_val = val;
	return true;
	}


//----------
//
// strcmplax--
//	Compare two strings, using lax case matching, and determine which precedes
//	the other alphabetically.
//
//----------
//
// Arguments:
//	const char*	s1:	One string.
//	const char*	s2:	The other string.
//
// Returns:
//	anything < 0 => s1 precedes s2 alphabetically.
//	0            => s1 and s2 are identical, except for possible case
//	                .. mismatches.
//	anything > 0 => s2 precedes s1 alphabetically.
//
//----------

int strcmplax
   (const char*	s1,
	const char*	s2)
	{
	char		c1, c2;

	do
		{
		c1 = *(s1++); c1 = tolower (c1);
		c2 = *(s2++); c2 = tolower (c2);
		} while ((c1 == c2) && (c1 != 0));

	return c1 - c2;
	}

//----------
//
// strncmplax--
//	Compare the prefixes of two strings, using lax case matching, and
//	determine which precedes the other alphabetically.
//
//----------
//
// Arguments:
//	const char*	s1:	One string.
//	const char*	s2:	The other string.
//	size_t		n:	The length of the prefixes to compare.
//
// Returns:
//	anything < 0 => s1's n-character prefix precedes s2's alphabetically.
//	0            => s1 and s2 have identical n-character prefixes, except for
//	                .. possible case mismatches.
//	anything > 0 => s2's n-character prefix precedes s1's alphabetically.
//
//----------

int strncmplax
   (const char*	s1,
	const char*	s2,
	size_t		n)
	{
	char		c1, c2;

	if (n <= 0)
		return 0;

	do
		{
		c1 = *(s1++); c1 = tolower (c1);
		c2 = *(s2++); c2 = tolower (c2);
		} while ((c1 == c2) && (c1 != 0) && (--n > 0));

	return c1 - c2;
	}

//----------
//
// new_string_f--
//  Allocate a string from the heap, using printf-style format to fill it.
//
//----------
//
// Arguments:
//  (as per printf)
//
// Returns:
//  A pointer to the string (s), or NULL if there was a failure.
//
//----------

char* new_string_f
   (const char* fmt, ...)
	{
	va_list		args;
	int			len;
	char*		s;

	va_start (args, fmt);

	len = vsnprintf (NULL, 0, fmt, args);
	s = ckalloc (len + 2); /* (probably should be len+1) */
	len = vsnprintf (s, len+1, fmt, args);

	va_end (args);
	return s;
	}

//----------
//
// chomp_gets--
//  Read a string from a stream.  This is identical to the standard C routine
//  fgets, except that this routine will not leave a newline at the end of the
//  returned string.
//
//----------
//
// Arguments:
//  char*   s:      The buffer to read the string into.
//  int     n:      The number of characters allowed in s, including a
//                  .. terminator.
//  FILE*   stream: The stream to read the string from.
//
// Returns:
//  A pointer to the string (s), or NULL if there was a failure reading the
//  stream.  Note that encountering an EOF is not considered a failure.
//
//----------

char* chomp_gets
   (char*	s,
    int		n,
    FILE*	stream)
    {
	s[0] = 0;

    if (fgets (s, n, stream) == NULL)
        return NULL;

	return chomp_string (s);
    }

//----------
//
// chomp_string--
//  Remove a newline from the end of a string.
//
//----------
//
// Arguments:
//  char*	s:	The buffer to read the string into.
//
// Returns:
//  A pointer to the string (s).
//
//----------

char* chomp_string
   (char*	s)
    {
    int		sLen;

    sLen = strlen (s);
    if ((s[sLen-1] == '\n') || (s[sLen-1] == '\r'))
        s[sLen-1] = 0;

    return s;
    }

//----------
//
// unsigned_digits_of, digits_of--
//  Determine how many digits a number has, *without* commas.
//
//----------
//
// Arguments:
//  long v: The number to represent.
//
// Returns:
//	The number of digits needed by that number, *not* including commas nor
//	a minus sign.
//
//----------

int digits_of
   (long	v)
	{
	if (v < 0) return unsigned_digits_of (-v);
	      else return unsigned_digits_of (v);
	}

int unsigned_digits_of
   (unsigned long	v)
	{
	int				digits;

	digits = 1;
	for ( ; v>9 ; v/=10)
		digits++;

	return digits;
	}

//----------
//
// next_lex_combo--
//  Generate (over the course of several calls) all vectors of N ones in M
//	bits, in lexicographical order.
//
// example usage:
// ==============
//
//	bitVector = 0;
//	while (true)
//		{
//		bitVector = next_lex_combo (numBits, numOnes, bitVector);
//		if (bitVector == 0)
//			break;
//
//		... use bitVector for something ...
//		}
//
// example bit order (for m=5, n=3):
// =================================
//
//	first: 1 1 1 0 0
//	       1 1 0 1 0
//	       1 1 0 0 1
//	       1 0 1 1 0
//	       1 0 1 0 1
//	       1 0 0 1 1
//	       0 1 1 1 0
//	       0 1 1 0 1
//	       0 1 0 1 1
//	 last: 0 0 1 1 1
//
//----------
//
// Arguments:
//	unsigned long	m:	The number of bits in each vector.
//	unsigned long	n:	The number of ones in each vector.  This must be
//						non-zero.
//	unsigned long	v:	The current vector;  0 indicates that we should
//						.. produce the first vector in the sequence.
//
// Returns:
//	The next vector;  0 if there are no more such vectors.
//
// Warning:
//	This makes use of some serious bit twiddling.
//
//----------

unsigned long next_lex_combo
   (unsigned long	m,
	unsigned long	n,
	unsigned long	v)
	{
	static unsigned long	f;
	unsigned long			x, y, z, b;

	if (v == 0)
		{
		v = (1L << m) - (1L << (m-n));	// generate first state 111...000
		f = (1L << n) - 1;				// generate final state 000...111

		return v;
		}

	// generate the next state

	if ((v&1) == 0)						// is rightmost bit zero?
		{								// (yes, rightmost bit is zero)
		z = v&(-v);						// isolate rightmost one
		v -= z>>1;						// move it one to the right
		}
	else if (v == f)					// did we do the final state?
		return 0;						// (yes)
	else
		{								// (rightmost is one, not final)
		x = v&(v+1);					// clear all the rightmost ones
		y = x&(-x);						// isolate the one we want

		z = y>>2;						// position selector
		for (b=v>>1 ; (b&1)!=0; b>>=1)	// do for each rightmost one
			z>>=1;						//    move selector right

		v = x-z;						// add run of ones, order-flipped
		}

	return v;
	}

//----------
//
// str_to_chromosome--
//	Convert a string to an encoded chromosome.
//
//----------
//
// Arguments:
//	char*	s:	The string to convert.  This should be in the format
//				.. chr<number> or chr<letter>, where <number> is in the
//				.. range 1..99, and <letter> is upper case.  If the string
//				.. doesn't terminate at that point, it must contain
//				.. whitespace or punctuation immediately after the
//				.. chromosome.
//
// Returns:
//	The encoded chromosome;  0 if the string was not of the right format.
//	Numeric chromosomes are encoded as the number (up to 99).  Alphabetical
//	chromosomes are encoded as 100 ('A') thru 125 ('Z').  The value returned
//  will always be in the range 0..127 (so, for example, callers can create
//  an array of 128 entries to do something with the results).
//
//----------

static int isOk (char ch);
static int isOk (char ch)
	{ return ((ch==0) || (isspace(ch)) || (ispunct(ch))); }

char str_to_chromosome
   (char*	s)
	{
	if (s == NULL) return 0;

	if (strncmp (s, "chr", 3) != 0) return 0;
	s += 3;

	if (isupper (s[0]))
		{
		if (isOk(s[1])) return 100 + s[0]-'A';
		}
	else if (isdigit (s[0]))
		{
		if (isOk(s[1])) return s[0]-'0';
		if (isdigit (s[1]))
			{
			if (isOk(s[2])) return 10*(s[0]-'0') + (s[1]-'0');
			}
		}

	return 0;
	}

//----------
//
// chromosome_to_str--
//	Convert an encoded chromosome to a string.
//
//----------
//
// Arguments:
//  char	chr:	The chromosome to convert.  This should be encoded as per
//					str_to_chromosome();
//
// Returns:
//  A string containing the representation of the chromosome, not longer than
//	five characters.  This string is actually static data belonging to this
//	routine, so the caller must copy it if more than one converted chromosome
//	is to be used simultaneously.
//
//----------

char* chromosome_to_str
   (char	chr)
	{
	static	char s[6];

	strcpy (s, "chr");

	if      (chr <   0)  strcpy  (s+3, "**");
	else if (chr <=  99) sprintf (s+3, "%d", chr);
	else if (chr <= 125) sprintf (s+3, "%c", 'A' + chr-100);
	else                 strcpy  (s+3, "**");

	return s;
	}

//----------
//
// nuc3_to_aa--
//	Convert a nucleotide triplet to an amino acid, using the standard genetic
//	code.
//
//----------
//
// Arguments:
//  unsigned char*	nuc:	The nucleotide triplet to convert.  This is an
//							.. array of at least three characters, each of
//							.. which is A, C, G, T, or a, c, g, t.
//
// Returns:
//	The corresponding amino acid character (upper case).  If the nucleotide
//	triplet was bad, '?' is returned.
//
//----------

#define _ (-1)

const signed char nuc_to_bits[256] =
	{
	_, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _,	// 0x
	_, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _,	// 1x
	_, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _,	// 2x
	_, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _,	// 3x (numbers)
	_, 2, _, 1, _, _, _, 3, _, _, _, _, _, _, _, _,	// 4x (upper case)
	_, _, _, _, 0, _, _, _, _, _, _, _, _, _, _, _,	// 5x (upper case)
	_, 2, _, 1, _, _, _, 3, _, _, _, _, _, _, _, _,	// 6x (lower case)
	_, _, _, _, 0, _, _, _, _, _, _, _, _, _, _, _,	// 7x (lower case)
	_, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _,	// 8x
	_, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _,	// 9x
	_, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _,	// Ax
	_, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _,	// Bx
	_, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _,	// Cx
	_, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _,	// Dx
	_, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _,	// Ex
	_, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _	// Fx
	};

const char* bits_to_nuc = "TCAG";

const char* bits_to_aa =
//	 TTTTTTTTTTTTTTTT   CCCCCCCCCCCCCCCC   AAAAAAAAAAAAAAAA   GGGGGGGGGGGGGGGG
//	 TTTTCCCCAAAAGGGG   TTTTCCCCAAAAGGGG   TTTTCCCCAAAAGGGG   TTTTCCCCAAAAGGGG
//	 TCAGTCAGTCAGTCAG   TCAGTCAGTCAGTCAG   TCAGTCAGTCAGTCAG   TCAGTCAGTCAGTCAG
	"FFLLSSSSYY**CC*W" "LLLLPPPPHHQQRRRR" "IIIMTTTTNNKKSSRR" "VVVVAAAADDEEGGGG";

char nuc3_to_aa
   (unsigned char*	nuc)
	{
	int				aaCode;

	aaCode = nuc3_to_bits (nuc);

	if (aaCode < 0) return '?';
	           else return bits_to_aa[aaCode];
	}

//----------
//
// nuc3_to_bits--
//	Convert a nucleotide triplet to a six-bit code.
//
// The code assigns two bits for each nucleotide, encoding T as 00, C as 01,
// A as 10, and G as 11.  The third nucleotide is placed in the least
// significant bits.
//
// For example,
//
//		CAT ==> 01 10 00 ==> 011000 ==> (decimal) 24
//
//----------
//
// Arguments:
//  unsigned char*	nuc:	The nucleotide triplet to convert.  This is an
//							.. array of at least three characters, each of
//							.. which is A, C, G, T, or a, c, g, t.
//
// Returns:
//	The corresponding six-bit code.  If the nucleotide triplet was bad, -1 is
//	returned.
//
//----------

char nuc3_to_bits
   (unsigned char*	nuc)
	{
	int				aaCode;
	int				bits;

	aaCode = 0;
	if ((bits = nuc_to_bits[nuc[0]]) < 0) return -1;  aaCode += bits << 4;
	if ((bits = nuc_to_bits[nuc[1]]) < 0) return -1;  aaCode += bits << 2;
	if ((bits = nuc_to_bits[nuc[2]]) < 0) return -1;  aaCode += bits;

	return aaCode;
	}

//----------
//
// is_wild_match--
//	Determine if a string matches a pattern, where the pattern may contain
//	wildcards.
//
// Wildcards are represented by either a . (match any single character) or *
// (match a string of characters of any length, including zero).  For example,
// ab.*c* match abbc, abwxyzcwwwww, and so on.  \ can be used as an escape
// character to get * (\*), . (\.), or \ (\\) into the pattern.  A \ at the
// end of the string is treated as no character.
//
//----------
//
// The algorithm used examines characters in p (the pattern) one at a time,
// forming the set of prefixes of s (the test string) that match the prefix of
// p.  Formally, this set is defined as follows
//
//	m[j] = { i | s[0..i] matches p[0..j] }
//
// where, for example, p[0..j] is the prefix of p consisting of characters 0
// thru j.  A recurrence relation for m is
//
//	m[-1] = { -1 }
//
//	m[j] = if m[j] == '*' { i+k | i-1 is in m[j-1]  and  0<=k<=(|s|-1)-(i-1) }
//	       otherwise      { i   | i-1 is in m[j-1]  and  s[i] matches p[j] }
//
// We have an overall match if and only if |s|-1 is in m[|p|-1].
//
//----------
//
// Arguments:
//	const char*	s:			The string to test.
//	const char*	pattern:	The pattern to match against, which may contain
//							.. wildcards.
//
// Returns:
//	true if we have a match;  false if not.  If the routine fails (due to
//	memory allocation failure), we also return false;
//
//----------

static int is_match (const char s, const char p);
static int is_match (const char s, const char p)
	{ return (tolower (s) == tolower (p)); }

// is_wild_match--

int is_wild_match
   (const char*	s,
	const char*	pattern)
	{
	char*		m;
	char*		mbig = NULL;
	char		mlittle[100];
	int 	    sLen = strlen (s);
	int			i;
	int			anyMatches, haveMatch;

	//////////
	// initialize the m-set
	//////////

	// decide which memory to use;  allocate memory if needed;  note that m
	// is indexed by -1..sLen-1

	if (1+sLen <= sizeof(mlittle))
		m = mlittle + 1;
	else
		{
		mbig = (char*) malloc ((1+sLen) * sizeof(char));
		if (mbig == NULL) return false;
		m = mbig + 1;
		}

	// initialize the m-set

	m[-1] = true;

	for (i=0 ; i<sLen ; i++)
		m[i] = false;

	//////////
	// scan the pattern;  for each character in the pattern, we compute the
	// new m-set from the old one
	//////////

	for ( ; *pattern!=0 ; pattern++)
		{
		switch (*pattern)
			{
			// escape character-- advance to next character and fall through
			// to ordinary character case

			case '\\':
				pattern++;
				if (*pattern == 0) continue;
				// fall through;

			// ordinary character-- try to extend all matches by one character

			default:
				for (i=sLen-1 ; i>=0 ; i--)
					{
					if (m[i-1])								// i-1 is in m[j-1]
						m[i] = is_match (s[i], *pattern);	// s[i] matches p[j]
					else
						m[i] = false;
					}
				m[-1] = false;
				break;

			// single wild character-- extend all matches by one character

			case '.':
				for (i=sLen-1 ; i>=0 ; i--)
					m[i] = m[i-1];
				m[-1] = false;
				break;

			// wild substring-- everything as long as the shortest match so
			// far is a match

			case '*':
				for (i=-1 ; i<sLen ; i++)
					{ if (m[i]) break; }
				for ( ; i<sLen ; i++)
					m[i] = true;
				break;
			}

		// check to see if the m-set is empty, which would mean we have no
		// chance at a match

		anyMatches = false;

		for (i=-1 ; i<sLen ; i++)
				if (m[i])
			{ anyMatches = true;  break; }

		if (!anyMatches)
			break;
		}

	//////////
	// determine if we have a match
	//////////

	haveMatch = m[sLen-1];

	// free allocated space

	if (mbig != NULL)
		free (mbig);

	// tell the caller whether or not we have a match

	return haveMatch;
	}

//----------
//
// bit_count--
//	Count the '1' bits in a long value.
//
// Note:
//	- this algorithm was adapated from one written by Glenn C. Rhoads
//	  <rhoads470@my-deja.com> of the Computer Science Deptartment at Rutgers.
//
//----------
//
// Arguments:
//	unsigned long	bits:	The value to count the '1' bits of.
//
// Returns:
//	The number of bits that are '1'.
//
// Warning:
//	- assumes a long holds no more than 128 bits.
//
//----------

int bit_count
   (unsigned long	bits)
	{
	const unsigned long allBits      = ~0L;
	const unsigned long mask10       = (allBits/ 3) << 1;
	const unsigned long mask0011     =  allBits/ 5;
	const unsigned long mask00001111 =  allBits/17;
	int shift;

	// convert each pair to a count in the range 0..2
	//  00 => 00
	//  01 => 01
	//  10 => 01
	//  11 => 10

	bits -= (bits & mask10) >> 1;

	// convert each nybble to a count in the range 0..4

	bits = (bits & mask0011) + ((bits>>2) & mask0011);

	// convert each byte to a count in the range 0..8

	bits = (bits + (bits >> 4)) & mask00001111;

	// sum counts over bytes, then 16-bit words, and so on

	bits += bits >> 8;
	bits += bits >> 16;

	shift = 32;
	while (shift < 8*sizeof(unsigned long))
		{ bits += bits >> shift;  shift *= 2; }

	return bits & 0x000000FF;
	}


//----------
//
// greatest_common_divisor--
//  Find the greatest common divisor of two numbers.
// least_common_multiple--
//  Find the least common multiple of two numbers.
//
//----------
//
// Arguments:
//	int	a, b:	The numbers.  Note that this routine only deals with absolute
//				values.  See the warning below. 
//
// Returns:
//	(for gcd) The largest number that is a divisor of both a and b.
//	(for lcm) The smallest number that is a multiple of both a and b.
//
// Warnings:
//	- Negative numbers are converted to their absolute value.  But the most
//	  negative number (for example -2^31 if you have 32-bit ints) will not be
//	  handled correctly, since it's absolute value is not representable.  This
//	  routine does not check for that value, and will give erroneous results
//	  if that value is passed to it.
//	- Overflow is not checked.  If the least common multiple is larger than
//	  the largest integer, erroneous results are returned.  However, care is
//	  taken so that if the mathematical lcm is not larger than that, overflow
//	  will not occur.
//
//----------

int greatest_common_divisor
   (int		a,
	int		b)
	{
	int		r;

	// we deal only with the absolute values

	if (a < 0) a = -a;
	if (b < 0) b = -b;

	// make sure a is the larger of the two

	if (a < b)
		{ r = a; a = b; b = r; }

	// make sure we're not being asked to do something silly

	if (b == 0)
		return 0;

	// do it

	while (true)
		{
		r = a % b;
		if (r == 0) break;
		a = b;
		b = r;
		}

	return b;
	}

int least_common_multiple
   (int		a,
	int		b)
	{
	int		g;

	// we deal only with the absolute values

	if (a < 0) a = -a;
	if (b < 0) b = -b;

	// compute the gcd

	g = greatest_common_divisor (a, b);

	if (g == 0) return 0;

	// from that, compute the lcm

	return (a / g) * b;
	}

