//-------+---------+---------+---------+---------+---------+---------+--------=
//
// File: lector.c									Author: Bob Harris
//
// Generic C application program
//
//----------

//----------
//
// lector--
//  Given a list of genome ranges ("selectors") and a list of feature ranges
//	("features"), divide the features into two sets based on the amount of
//	overlap with elements in the selector set.
//
// $$$ Currently this reads all features into memory.  For large feature sets
// $$$ .. this might be a problem;  if allocation fails the program simply
// $$$ .. dies.  The program could be improved by ceasing to read features at
// $$$ .. the point where it runs out of memory.  These features can be
// $$$ .. processed, then it could free all that memory and pick up where it
// $$$ .. left off.  The difficulty is that it needs to have enough memory
// $$$ .. arround to read the selectors one by one.  So perhaps the selector
// $$$ .. should be read into pre-allocated memory rather than dynamically-
// $$$ .. allocated.
//----------

static const char rcsid[]=
"";

//----------
//
// other files--
//
//----------

#include <stdlib.h>				// standard C stuff
#define  true  1
#define  false 0
#include <stdio.h>				// standard C i/o stuff
#include <ctype.h>				// standard C upper/lower case stuff
#include <limits.h>				// standard C value limit stuff
#include "arg.h"				// simple argument processor
#include "ranges.h"				// Bob's range list stuff
#include "get_line.h"			// Scott Schwartz's line reading stuff
#include "util.h"				// Scott Schwartz's handy utility stuff
#include "bobutil.h"			// Bob's handy utility stuff

//----------
//
// global data--
//
//----------

typedef unsigned long  chunk;
#define chunkBits      (8*sizeof(chunk))
#define chunkAll       ((chunk) -1)
#define chunkRight(b)  ((chunk)((1ULL << (b)) - 1))

typedef struct
	{
	int		len;				// number of entries in val[]
	int		val[1];				// (variable length)
	} vector;

// options

int verbosity       = 0;		// how much info to bombard the user with
								//   0  => only the essentials
								//   10 => everything
								//   20 => private debug features
int overlapThresh   = 1;		// number of bp that a feature must have in
int overlapDenom	= 0;		// .. common with union of selectors to be
								// .. considered to 'overlap';  if overlapDenom
								// .. is non-zero, overlapThresh/overlapDenom
								// .. is a fraction of the feature's length
int chooseSelectors = false;	// true  => features select selectors
								// false => (normal) selectors select features

// selectors

char*		selectorsName=NULL;	// name of selectors file

// features

char*		featuresName = NULL;// name of features file
ranges_t	features;

int			chromTo1stIx[128];  // map chromosome number (as per range_t.chr)
								// .. to index into features list

// output files

int			countPositives;
int			countNegatives;

char		positiveMark;		// marks to append to postives and/or negatives
char		negativeMark;		// .. on output;  zero disables these

int			sendToStdout;		// true => send all positives and negatives to
								//         .. stdout
char*		positivesName=NULL;	// name of positives file
char*		negativesName=NULL;	// name of negatives file
char*		distribName=NULL;	// name of positions distribution file

//----------
//
// prototypes for local functions--
//
//----------

int main (int argc, char** argv);

static void usage              (void);
static void detail             (void);
static void get_args           (int argc, char** argv,
                                char** selectorsName, char** featuresName,
                                char** positivesName, char** negativesName,
                                char** distribName);
static void show_options       (FILE* f);
static int  string_to_overlap  (char* s, int* numerator, int* denominator);
static void sort_features      (ranges_t* r);
static void alloc_feature_bits (ranges_t* r, vector** posDistrib);
static void scan_selectors     (FILE* selF, ranges_t* r, vector* posDistrib);
static void choose_selectors   (FILE* selF, ranges_t* r,
                                FILE* posF, FILE* negF);
static int  count_overlap      (range_t* feat,
                                unsigned long start, unsigned long end);
static int  record_overlap     (range_t* feat, vector* posDistrib,
                                unsigned long start, unsigned long end);
static void divide_features    (ranges_t* r, FILE* posF, FILE* negF);
static int  enough_overlap     (int overlap, int featureLen);
static void print_distribution (vector* posDistrib, FILE* distF);

//----------
//
// lector--
//  Main program
//
//----------

int main
   (int			argc,
	char**  	argv)
	{
	FILE*		featuresF  = NULL;
	FILE*		selectorsF = NULL;
	FILE*		positivesF = NULL;
	FILE*		negativesF = NULL;
	FILE*		distribF   = NULL;
	vector*		posDistrib = NULL;
	ranges_t	emptyRange = ranges_INIT (ckrealloc, ckfree);

	//////////
	// fetch arguments
	//////////

	get_args (argc, argv, &selectorsName, &featuresName,
	                      &positivesName, &negativesName, &distribName);

	if (verbosity >= 8)
		show_options (stderr);

	//////////
	// open the input files
	//////////

	// open and read in the features

	features = emptyRange;

	if (featuresName == NULL)
		ranges_read (stdin, &features);
	else
		{
		featuresF = ckopen (featuresName, "rt");
		ranges_read (featuresF, &features);
		fclose (featuresF);
		}

	// sort them, and record index range for each chromosome

	sort_features (&features);

	if (verbosity >= 9)
		{
		printf ("features\n");
		printf ("-------\n");
		ranges_fprintf (stdout, &features, "%s %lu %lu %s %s\n");
		printf ("\n");
		}

	else if (verbosity >= 8)
		fprintf (stderr, "(%d features)\n", features.len);

	// open the selectors

	selectorsF = ckopen (selectorsName, "rt");

	// open the output files

	if (sendToStdout)
		positivesF = negativesF = stdout;
	else
		{
		if (positivesName != NULL)
			positivesF = ckopen (positivesName, "wt");
		else if (countPositives || countNegatives)
			positivesF = NULL;
		else
			positivesF = stdout;
	
		if (negativesName != NULL)
			negativesF = ckopen (negativesName, "wt");
		else
			negativesF = NULL;
		}

	if (distribName != NULL)
		distribF = ckopen (distribName, "wt");
	else
		distribF = NULL;

	//////////
	// allocate bit vectors for each feature
	//////////

	alloc_feature_bits (&features, (distribF!=NULL)? &posDistrib : NULL);

	//////////
	// accumulate the overlap stats
	//////////

	if (chooseSelectors)
		choose_selectors (selectorsF, &features, positivesF, negativesF);
	else
		scan_selectors (selectorsF, &features, posDistrib);

	//////////
	// print the features, divided;  note that if we're choosing selectors,
	// they have already been output
	//////////

	if (!chooseSelectors)
		divide_features (&features, positivesF, negativesF);

	if (posDistrib != NULL)
		print_distribution (posDistrib, distribF);

	//////////
	// clean up
	//////////

	ckfree (selectorsName);
	ckfree (featuresName);
	ckfree (positivesName);
	ckfree (negativesName);
	ckfree (distribName);

	if  (selectorsF != NULL)                            fclose (selectorsF);
	if ((positivesF != stdout) && (positivesF != NULL)) fclose (positivesF);
	if ((negativesF != stdout) && (negativesF != NULL)) fclose (negativesF);
	if  (distribF   != NULL)                            fclose (distribF);

	ckfree (posDistrib);

	ranges_fini (&features);

	return 0;
	}

//----------
//
// get_args--
//  Parse command line arguments.
//
//----------
//
// Arguments:
//	argc, argv:				(as per main)
//	char**	selectorsName:	Place to return the name of the selectors file.
//	char**	featuresName:	Place to return the name of the features file.  If
//							.. the user hasn't specified a file, NULL is
//							.. returned.
//	char**	positivesName:	Place to return the name of the positives file.
//							.. If the user hasn't specified a file, NULL is
//							.. returned.
//	char**	negativesName:	Place to return the name of the negatives file.
//							.. If the user hasn't specified a file, NULL is
//							.. returned.
//	char**	distribName:	Place to return the name of the positions
//							.. distribution file.  If the user hasn't
//							.. specified a file, NULL is returned.
//
// Returns:
//  (nothing)
//
// Globals Set:
//	All options and feature descriptors.
//
//----------

static void usage (void)
	{
	fprintf (stderr, "lector-- divide features into two sets based on overlap with a selector set\n");
	fprintf (stderr, "         features and selectors are lists of genomic intervals\n");
	fprintf (stderr, "usage: lector [options] selectors [features]\n");
	fprintf (stderr, "  [-detail]         print more detailed info about what this program does\n");
	fprintf (stderr, "  selectors         file containing a list of ranges to 'select' features\n");
	fprintf (stderr, "                    e.g. in one of these three formats\n");
	fprintf (stderr, "                      chr1 10241725 10241839 +\n");
	fprintf (stderr, "                      chr1:10241725-10241839 +\n");
	fprintf (stderr, "                      'chr1',10241725,10241839,'+'\n");
	fprintf (stderr, "                    note that ranges are origin 1, inclusive (like those used\n");
	fprintf (stderr, "                    in the UCSC genome browser)\n");
	fprintf (stderr, "  features          file(s) containing a list of ranges to divide\n");
	fprintf (stderr, "                    if features file not present, stdin is used\n");
	fprintf (stderr, "  [-p positives]    file to receive the features that overlap with selectors\n");
	fprintf (stderr, "                    if positives file not present, stdout is used\n");
	fprintf (stderr, "  [-n negatives]    file for features that don't overlap selectors\n");
	fprintf (stderr, "  [-stdout]         send both postives and negatives to stdout\n");
	fprintf (stderr, "                    (used in conjunction with -pn and -nm marks)\n");
	fprintf (stderr, "  [-pm pos_mark]    mark to append to positives\n");
	fprintf (stderr, "  [-nm neg_mark]    mark to append to negatives\n");
	fprintf (stderr, "  [-c overlap]      number of bp that a feature must have in common for overlap\n");
	fprintf (stderr, "                    (default is 1 bp)\n");
	fprintf (stderr, "  [-c numer/denom]  overlap threshold as a fraction of a feature's bp\n");
	fprintf (stderr, "  [-c pct%%]         overlap threshold as a percentage of a feature's bp\n");
	fprintf (stderr, "  [-countpos]       output a count of positives instead of the features\n");
	fprintf (stderr, "  [-countneg]       output a count of negatives instead of the features\n");
	fprintf (stderr, "  [-d distribution] file for position distribution of overlap\n");
// -viceversa is too complicated to describe correctly, so hide it
//	fprintf (stderr, "  [-viceversa]      output selectors, selected by features\n");
//	fprintf (stderr, "                    (use this when otherwise feature set would be huge)\n");
	fprintf (stderr, "  [-v verbosity]    set info level (0 is minimum, 10 is everything)\n");
	exit(1);
	}

static void detail (void)
	{
	fprintf (stderr, "(not yet complete)\n");
	fprintf (stderr, "Given a list of genome ranges (\"selectors\") and a list of feature ranges\n");
	fprintf (stderr, "(\"features\"), divide the features into two sets based on the amount of\n");
	fprintf (stderr, "overlap with elements in the selector set.  A feature is considered a\n");
	fprintf (stderr, "positive if it has enough positions that overlap, in total, with any of the\n");
	fprintf (stderr, "selectors.\n");
	fprintf (stderr, "\n");
//	fprintf (stderr, "Alternatively, the features can be used to divide the selector set.  This is\n");
//	fprintf (stderr, "enabled by the -viceversa option.  A selector is considered a positive if it\n");
//	fprintf (stderr, "has enough overlap with any single feature.\n");
//	fprintf (stderr, "\n");
	fprintf (stderr, "Overlap detection uses a bit vector for each feature, with a bit for each bp\n");
	fprintf (stderr, "in the feature.  Selectors are compared to the features, setting any bits they\n");
	fprintf (stderr, "overlap.  The number of set bits in the vector is compared to the overlap\n");
	fprintf (stderr, "threshold to determine positive or negative.\n");
	fprintf (stderr, "\n");
	fprintf (stderr, "The implementation thus reads all the features into memory, then reads through\n");
	fprintf (stderr, "the selectors one by one without committing them to memory.  Thus it is not\n");
	fprintf (stderr, "necessary to sort either set externally.\n");
	exit(1);
	}

static void get_args
   (int			_argc,
	char**		_argv,
	char**		selectorsName,
	char**		featuresName,
	char**		positivesName,
	char**		negativesName,
	char**		distribName)
	{
	int			argc;
	char**		argv;

	*selectorsName = NULL;
	*featuresName  = NULL;
	*positivesName = NULL;
	*negativesName = NULL;
	*distribName   = NULL;

	// skip program name

	argv = _argv+1;  argc = _argc - 1;

	//////////
	// scan arguments
	//////////

	sendToStdout   = false;
	countPositives = false;
	countNegatives = false;
	positiveMark   = 0;
	negativeMark   = 0;

	while (argc > 0)
		{
		// -detail

		if (strcmp (argv[0], "-detail") == 0)
			{
			detail ();
			argv++;  argc--;
			continue;
			}

		// -c overlap

		if ((argc >= 2) && (strcmp (argv[0], "-c") == 0))
			{
			if (!string_to_overlap (argv[1], &overlapThresh, &overlapDenom))
				{
				fprintf (stderr, "%s %s is not valid\n", argv[0], argv[1]);
				usage ();
				}
			argv+=2;  argc-=2;
			continue;
			}

		// -v verbosity

		if ((argc >= 2) && (strcmp (argv[0], "-v") == 0))
			{
			if (!string_to_int (argv[1], &verbosity))
				{
				fprintf (stderr, "%s %s is not valid\n", argv[0], argv[1]);
				usage ();
				}
			argv+=2;  argc-=2;
			continue;
			}

		// -pm pos_mark

		if ((argc >= 2) && (strcmp (argv[0], "-pm") == 0))
			{
			if (strlen (argv[1]) != 1)
				{
				fprintf (stderr, "%s %s is not valid\n", argv[0], argv[1]);
				usage ();
				}
			positiveMark = argv[1][0];
			argv+=2;  argc-=2;
			continue;
			}

		// -nm neg_mark

		if ((argc >= 2) && (strcmp (argv[0], "-nm") == 0))
			{
			if (strlen (argv[1]) != 1)
				{
				fprintf (stderr, "%s %s is not valid\n", argv[0], argv[1]);
				usage ();
				}
			negativeMark = argv[1][0];
			argv+=2;  argc-=2;
			continue;
			}

		// -stdout

		if (strcmp (argv[0], "-stdout") == 0)
			{
			if ((*positivesName != NULL) || (*negativesName != NULL))
				{
				fprintf (stderr, "can't use -stdout with -p or -n\n");
				usage ();
				}
			sendToStdout = true;
			argv++;  argc--;
			continue;
			}

		// -viceversa

		if (strcmp (argv[0], "-viceversa") == 0)
			{
			chooseSelectors = true;
			argv++;  argc--;
			continue;
			}

		// -countpos

		if (strcmp (argv[0], "-countpos") == 0)
			{
			countPositives = true;
			argv++;  argc--;
			continue;
			}

		// -countneg

		if (strcmp (argv[0], "-countneg") == 0)
			{
			countNegatives = true;
			argv++;  argc--;
			continue;
			}

		// postitives file name

		if ((*positivesName == NULL)
		 && (argc >= 2) && (strcmp (argv[0], "-p") == 0))
			{
			if (sendToStdout)
				{
				fprintf (stderr, "can't use -p with -stdout\n");
				usage ();
				}
			*positivesName = copy_string (argv[1]);
			argv+=2;  argc-=2;
			continue;
			}

		// negatives file name

		if ((*negativesName == NULL)
		 && (argc >= 2) && (strcmp (argv[0], "-n") == 0))
			{
			if (sendToStdout)
				{
				fprintf (stderr, "can't use -n with -stdout\n");
				usage ();
				}
			*negativesName = copy_string (argv[1]);
			argv+=2;  argc-=2;
			continue;
			}

		// distribution file name

		if ((*distribName == NULL)
		 && (argc >= 2) && (strcmp (argv[0], "-d") == 0))
			{
			*distribName = copy_string (argv[1]);
			argv+=2;  argc-=2;
			continue;
			}

		// selectors file name

		if (*selectorsName == NULL)
			{
			*selectorsName = copy_string (argv[0]);
			argv++;  argc--;
			continue;
			}

		// features file name

		if (*featuresName == NULL)
			{
			*featuresName = copy_string (argv[0]);
			argv++;  argc--;
			continue;
			}

		fprintf (stderr, "Can't understand \"%s\"\n", argv[0]);
		usage ();
		}

	//////////
	// check for sensibility
	//////////

	if (*selectorsName == NULL)
		{
		fprintf (stderr, "You must specify a selectors file\n");
		usage ();
		}

	if (overlapThresh <= 0)
		{
		fprintf (stderr, "Overlap must be strictly positive\n");
		usage ();
		}

	if ((overlapDenom != 0) && (overlapThresh > overlapDenom))
		{
		fprintf (stderr, "Overlap percentage cannot exceed 100%%\n");
		usage ();
		}

	if ((sendToStdout) && (positiveMark == 0) && (negativeMark == 0))
		{
		fprintf (stderr, "-stdout requires -pm or -nm\n");
		usage ();
		}

	if ((chooseSelectors) && (*distribName != NULL))
		{
		fprintf (stderr, "Cannot use -d with -viceversa\n");
		usage ();
		}

	}

//----------
//
// show_options--
//  Show the options that were set by the command line arguments.
//
//----------
//
// Arguments:
//  FILE*   f:  The stream to write to.
//
// Returns:
//  (nothing)
//
//----------

static void show_options
   (FILE*  f)
	{
	fprintf (f, "verbosity     = %d\n", verbosity);
	if (overlapDenom == 0)
		fprintf (f, "overlapThresh = %d\n", overlapThresh);
	else
		fprintf (f, "overlapThresh = %d/%d\n", overlapThresh, overlapDenom);
	fprintf (f, "selectorsName = \"%s\"\n", selectorsName);

	if (featuresName == NULL)
		fprintf (f, "featuresName  = (null)\n");
	else
		fprintf (f, "featuresName  = \"%s\"\n", featuresName);

	if (positivesName == NULL)
		fprintf (f, "positivesName = (null)\n");
	else
		fprintf (f, "positivesName = \"%s\"\n", positivesName);

	if (negativesName == NULL)
		fprintf (f, "negativesName = (null)\n");
	else
		fprintf (f, "negativesName = \"%s\"\n", negativesName);

	if (distribName == NULL)
		fprintf (f, "distribName   = (null)\n");
	else
		fprintf (f, "distribName   = \"%s\"\n", distribName);
	}

//----------
//
// string_to_overlap--
//  Convert a string to an overap threshold (integer or fraction)
//
//----------
//
// Arguments:
//	char*	s:			The string to convert.
//  int*	numerator:	Place to return the threshold numerator.  This will
//						.. only be written to if the conversion is successsful.
//  int*	denominator:Place to return the threshold denominator.  This will
//						.. only be written to if the conversion is successsful.
//
// Returns:
//  true if the string contained a valid overap threshold and nothing else;
//  false otherwise.
//
//----------

static int string_to_overlap
   (char*	s,
	int*	_numerator,
	int*	_denominator)
	{
	int		numerator, denominator;
	char*   ss = NULL;
	int		len;
	char*   slash;

	// if the string is NULL or empty, it's invalid

	if (s == NULL) return false;
	if (*s == 0)   return false;

	// if it finishes with a percent, treat it as a percentage

	len = strlen (s);

	if (s[len-1] == '%')
		{
		ss = copy_string (s);
		ss[len-1] = 0;
		if (!string_to_int (ss, &numerator)) return false;
		denominator = 100;
		}

	// if it contains a /, parse it as a fraction

	else if (strchr (s, '/') != NULL)
		{
		ss    = copy_string (s);
		slash = strchr (ss, '/');
		*(slash++) = 0;
		if (!string_to_int (ss,    &numerator))   return false;
		if (!string_to_int (slash, &denominator)) return false;
		if (denominator <= 0)                     return false;
		}

	// otherwise, parse it as a simple threshold value

	else
		{
		if (!string_to_int (s, &numerator)) return false;
		denominator = 0;
		}

	// success!

	if (ss != NULL) free (ss);

	*_numerator   = numerator;
	*_denominator = denominator;
	return true;
	}

//----------
//
// sort_features--
//  Sort the features, and record the first index of each chromosome.
//
//----------
//
// Arguments:
//	ranges_t*	r:	The list of feature ranges.
//
// Returns:
//  (nothing;  failure terminates program)
//
//----------

static void sort_features
   (ranges_t*	r)
	{
	range_t*	feat;
	int			prevChr;
	int			ix;

	// sort the features

	ranges_sort (&features, NULL);

	// find first index for each chromosome

	prevChr = -1;

	for (ix=0 ; ix<r->len ; ix++)
		{
		feat = &r->vec[ix];
		if (feat->chr == prevChr) continue;

		while (++prevChr < feat->chr)
			chromTo1stIx[prevChr] = ix;

		chromTo1stIx[(unsigned) feat->chr] = ix;
		prevChr = feat->chr;
		}

	while (++prevChr < 128)
		chromTo1stIx[prevChr] = ix;

	}

//----------
//
// alloc_feature_bits--
//  Allocate (and clear) a bit vector for every feature.
//
//----------
//
// Arguments:
//	ranges_t*	r:			The list of feature ranges.
//	vector**	posDistrib:	Place to return an allocated vector for the
//							.. positions distribution.  If the pointer is NULL,
//							.. no vector is allocated.
//
// Returns:
//  (nothing;  failure terminates program)
//
// Notes:
//  The bit vectors will be automatically deallocated when ranges_fini() is
//	called.
//
//----------

static void alloc_feature_bits
   (ranges_t*		r,
	vector**		posDistrib)
	{
	range_t*		feat;
	int				numBits, maxBits;
	int				ix;

	maxBits = 0;

	for (ix=0 ; ix<r->len ; ix++)
		{
		feat = &r->vec[ix];

		// determine number of bits needed, rounded up to a multiple of whole
		// longs

		numBits =  feat->end+1 - feat->start;
		if (numBits > maxBits) maxBits = numBits;

		numBits += chunkBits - 1;
		numBits -= numBits % chunkBits;

		// allocate and install bit vector, with all bits set to zero

		feat->info    = 0;
		feat->infoPtr = (chunk*) ckallocz (numBits / 8);
		}

	if (posDistrib != NULL)
		{
		int bytesNeeded = sizeof(vector) + (maxBits-1)*sizeof(int);
		*posDistrib = (vector*) ckallocz (bytesNeeded);
		(*posDistrib)->len = maxBits;
		}
	}

//----------
//
// scan_selectors--
//	Scan the list of selectors, comparing each to every feature (on the same
//  chromosome) and recording any bp's that overlap (in the feature's bit
//  vector).
//
// $$$ Performance could probably be improved by doing a binary search for
// $$$ .. features that match the selector, as opposed to scanning the whole
// $$$ .. chromosome.  This is not as simple as a standard binary search,
// $$$ .. since there may be several features that overlap the selector.
// $$$ .. The same improvement would also apply to choose_selectors.
//
//----------
//
// Arguments:
//	FILE*		selF:		The file containing the selectors.
//	ranges_t*	r:			The list of feature ranges.
//	vector*		posDistrib:	A vector of ints in which to accumulate the
//							.. positions distribution.  This can be NULL.
//
// Returns:
//  (nothing)
//
//----------

static void scan_selectors
   (FILE*			selF,
	ranges_t*		r,
	vector*			posDistrib)
	{
	char*			buffer    = NULL;
	unsigned long	bufferLen = 0;
	long			lineNumber, selNumber;
	range_t			sel;
	range_t*		feat;
	unsigned char   chr;
	int				firstIx, lastIx, ix;

	lineNumber = selNumber = 0;
	while (true)
		{
		// read the next line;  note that get_line does not allow us to
		// distinguish an error from an end-of-file

		if (get_line (&buffer, &bufferLen, selF) == (unsigned long) -1) break;
		chomp_string (buffer);

		lineNumber++;

		// parse the range;  if it's a comment, ignore it

		if (!parse_range (buffer, &sel, (int) lineNumber))
			continue;
		selNumber++;

		if (verbosity >= 10)
			range_fprintf (stdout, &sel, "%s %lu %lu %s %s\n");

		// determine the range of features we should compare this selector to

		chr = sel.chr;

		firstIx = chromTo1stIx[chr];
		if (sel.chr == 127) lastIx = r->len-1;
		               else lastIx = chromTo1stIx[chr+1]-1;

		// compare this selector to each feature, recording any overlap

		for (ix=firstIx ; ix<=lastIx ; ix++)
			{
			feat = &r->vec[ix];

			if ((sel.start > feat->end)			// (any overlap?)
			 || (sel.end   < feat->start))
				continue;						// (nope), skip it

			// record the amount of overlap

			if (verbosity >= 20)
				range_fprintf (stdout, feat, "  %s %lu %lu %s %s => ");

			record_overlap (feat, posDistrib, sel.start, sel.end);

			if (verbosity >= 20)
				printf ("\n");
			}

		if ((verbosity >= 9) && ((selNumber % 1000000) == 0))
			fprintf (stderr, "(have processed %ld selectors)\n", selNumber);

		// give back any additional storage that was allocated with this
		// selector

		range_fini (&sel);
		}

	if (verbosity >= 8)
		fprintf (stderr, "(%ld selectors)\n", selNumber);

	ckfree (buffer);
	}

//----------
//
// choose_selectors--
//	Scan the list of selectors, comparing each to every feature (on the same
//  chromosome) and calling the selector a positive if it has enough overlap
//  with any single feature.
//
// $$$ See note in scan_selectors for possible performance improvement.
//
//----------
//
// Arguments:
//	FILE*		selF:	The file containing the selectors.
//	ranges_t*	r:		The list of feature ranges.
//	FILE*		posF:	The file to receive the positives.  This can be NULL
//						.. if positives are of no interest.
//	FILE*		negF:	The file to receive the negatives.  This can be NULL
//						.. if negatives are of no interest.
//
// Returns:
//  (nothing)
//
//----------

static void choose_selectors
   (FILE*			selF,
	ranges_t*		r,
	FILE*			posF,
	FILE*			negF)
	{
	char*			buffer    = NULL;
	unsigned long	bufferLen = 0;
	long			lineNumber, selNumber;
	int				posCnt, negCnt;
	int				overlapBits, enoughOverlap;
	range_t			sel;
	range_t*		feat = NULL;
	int				featLen;
	unsigned char   chr;
	int				firstIx, lastIx, ix;

	lineNumber = selNumber = 0;
	posCnt     = negCnt    = 0;

	while (true)
		{
		// read the next line;  note that get_line does not allow us to
		// distinguish an error from an end-of-file

		if (get_line (&buffer, &bufferLen, selF) == (unsigned long) -1) break;
		chomp_string (buffer);

		lineNumber++;

		// parse the range;  if it's a comment, ignore it

		if (!parse_range (buffer, &sel, (int) lineNumber))
			continue;
		selNumber++;

		if (verbosity >= 10)
			range_fprintf (stdout, &sel, "%s %lu %lu %s %s\n");

		// determine the range of features we should compare this selector to

		chr = sel.chr;

		firstIx = chromTo1stIx[chr];
		if (sel.chr == 127) lastIx = r->len-1;
		               else lastIx = chromTo1stIx[chr+1]-1;

		// compare this selector to each feature, determining if it has
		// 'enough' overlap with any single feature

		enoughOverlap = false;

		for (ix=firstIx ; ix<=lastIx ; ix++)
			{
			feat    = &r->vec[ix];
			featLen = feat->end+1 - feat->start;

			if ((sel.start > feat->end)			// (any overlap?)
			 || (sel.end   < feat->start))
				continue;						// (nope), skip it

			// record the amount of overlap

			if (verbosity >= 20)
				range_fprintf (stdout, feat, "  %s %lu %lu %s %s => ");

			overlapBits = count_overlap (feat, sel.start, sel.end);

			if (verbosity >= 20)
				printf ("\n");

			// see if we have enough

			if (enough_overlap (overlapBits, featLen))
				{ enoughOverlap = true;  break; }
			}

		// output this selector as either a positive or negative

		if (enoughOverlap)
			{
			if (posF != NULL) range_fprintf (posF, &sel, "%s %lu %lu %s %s\n");
			posCnt++;
			if (verbosity >= 20) printf ("  positive\n");
			}
		else
			{
			if (negF != NULL) range_fprintf (negF, &sel, "%s %lu %lu %s %s\n");
			negCnt++;
			if (verbosity >= 20) printf ("  negative\n");
			}

		if ((verbosity >= 9) && ((selNumber % 1000000) == 0))
			fprintf (stderr, "(have processed %ld selectors)\n", selNumber);

		// give back any additional storage that was allocated with this
		// selector

		range_fini (&sel);
		}

	if (verbosity >= 8)
		{
		fprintf (stderr, "(%ld selectors)\n", selNumber);
		fprintf (stderr, "(%d positives and %d negatives)\n",
		                 posCnt, negCnt);
		}

	ckfree (buffer);
	}

//----------
//
// count_overlap--
//	Count the amount of overlap between a region and a feature.
//
//----------
//
// Arguments:
//	range_t*		feat:	The feature for which overlap has been discovered.
//	unsigned long	start:	First position in overlapping region.  See note
//							.. below.
//	unsigned long	end:	Last position in overlapping region (inclusive).
//
// Returns:
//  The number of overlapping bits between the overlapping region and the
//  feature.
//
// Note:
//	- The start and end positions may extend beyond the actual range of the
//	  .. bit vector.  This routine clips the overlap to the feature's range.
//	- This routine does *not* record the bits in the feature's bit vector.
//    .. For that, call record_overlap().
//
//----------

static int count_overlap
   (range_t*		feat,
	unsigned long	start,
	unsigned long	end)
	{
	// trim endpoints and re-origin them to the feature

	if (start < feat->start) start = feat->start;
	if (end   > feat->end)   end   = feat->end;

	if (start > end)	// (ignore empty intervals)
		return 0;

	start -= feat->start;
	end   -= feat->start;

	if (verbosity >= 20)
		printf ("overlap: %d bits", (int) ((end+1) - start));

	return (end+1) - start;
	}

//----------
//
// record_overlap--
//	Record some overlap in a feature's bit vector.
//
//----------
//
// Arguments:
//	range_t*		feat:		The feature for which overlap has been
//								.. discovered.
//	vector*			posDistrib:	A vector of ints in which to accumulate the
//								.. positions distribution.  This can be NULL.
//	unsigned long	start:		First position in overlapping region.  See note
//								.. below.
//	unsigned long	end:		Last position in overlapping region (inclusive).
//
// Returns:
//  The number of overlapping bits between the overlapping region and the
//  feature.
//
// Note:
//	- The start and end positions may extend beyond the actual range of the
//	  .. bit vector.  This routine clips the overlap to the feature's range.
//	- We record the bits of overlap in the bits vector, and at the same time
//	  .. keep a count of any new bits that are set.
//
//----------

static int record_overlap
   (range_t*		feat,
	vector*			posDistrib,
	unsigned long	start,
	unsigned long	end)
	{
	int				chunkIx, leftZeros, rightZeros;
	unsigned long	nextStart;
	chunk*			bitVec;
	chunk			bitsToSet, newBits;
	int				newBitCount, allBitCount;

	// trim endpoints and re-origin them to the bit vector

	if (start < feat->start) start = feat->start;
	if (end   > feat->end)   end   = feat->end;

	if (start > end)	// (ignore empty intervals)
		return 0;

	start -= feat->start;
	end   -= feat->start;

	// set the bits

	bitVec = (chunk*) feat->infoPtr;
	newBitCount = 0;
	allBitCount = 0;

	while (start <= end)
		{
		chunkIx   = start / chunkBits;
		leftZeros = start % chunkBits;
		nextStart = (chunkIx+1) * chunkBits;

		if (end >= nextStart - 1)
			rightZeros = 0;
		else
			rightZeros = (chunkBits-1) - (end % chunkBits);

		if (leftZeros == 0) bitsToSet =  chunkAll;
		               else bitsToSet =  chunkRight(chunkBits-leftZeros);
		if (rightZeros > 0) bitsToSet -= chunkRight(rightZeros);

		newBits = bitsToSet & ~bitVec[chunkIx];
		bitVec[chunkIx] |= newBits;
		newBitCount += bit_count (newBits);
		allBitCount += bit_count (bitsToSet);

		if ((posDistrib != NULL) && (newBits != 0))
			{
			int pos, strandPos;
			for (pos=(chunkIx*chunkBits)+(chunkBits-1) ;
			     newBits!=0                            ;
			     newBits>>=1, pos--)
					if ((newBits & 1) != 0)
				{
				strandPos = pos;
				if (feat->strand == '-')
					strandPos = (feat->end - feat->start) - pos;
				if (strandPos < posDistrib->len)
					{
					if (verbosity >= 20)
						printf ("overlap@ %d\n", strandPos);
					posDistrib->val[strandPos]++;
					}
				else
					{
					if (verbosity >= 20)
						printf ("overlap@ %d discarded\n", strandPos);
					}
				}
			}

		start = nextStart;
		}

	feat->info += newBitCount;

	if (verbosity >= 20)
		printf ("overlap: %d bits (%d new)", allBitCount, newBitCount);

	return allBitCount;
	}

//----------
//
// divide_features--
//	Divide the features set into two sets, based on the amount of overlap that
//	has been recorded in the feature bit vectors.
//
//----------
//
// Arguments:
//	ranges_t*	t:		The list of feature ranges.
//	FILE*		posF:	The file to receive the positives.  This can be NULL
//						.. if positives are of no interest.
//	FILE*		negF:	The file to receive the negatives.  This can be NULL
//						.. if negatives are of no interest.
//
// Returns:
//  (nothing)
//
//----------

static void divide_features
   (ranges_t*	r,
	FILE*		posF,
	FILE*		negF)
	{
	range_t*	feat;
	int			featLen;
	int			posCnt, negCnt;
	int			ix;

	if (verbosity >= 10)
		printf ("\ntotal overlap\n");

	posCnt = negCnt = 0;

	for (ix=0 ; ix<r->len ; ix++)
		{
		feat    = &r->vec[ix];
		featLen = feat->end+1 - feat->start;

		if (enough_overlap (feat->info, featLen))
			{
			if (posF != NULL)
				{
				range_fprintf (posF, feat, "%s %lu %lu %s %s");
				if (positiveMark != 0) fprintf (posF, "%c", positiveMark);
				fprintf (posF, "\n");
				}
			posCnt++;
			}
		else
			{
			if (negF != NULL)
				{
				range_fprintf (negF, feat, "%s %lu %lu %s %s");
				if (negativeMark != 0) fprintf (negF, "%c", negativeMark);
				fprintf (negF, "\n");
				}
			negCnt++;
			}

		if (verbosity >= 10)
			{
			printf        ("%ld (%0.1f%%) => ",
			               feat->info, (100.0*feat->info)/featLen);
			range_fprintf (stdout, feat, "%s %lu %lu %s %s\n");
			}
		}

	if (verbosity >= 8)
		fprintf (stderr, "(%d positives and %d negatives)\n",
		                 posCnt, negCnt);

	if (countPositives && countNegatives)
		printf ("%d %d\n", posCnt, negCnt);
	else if (countPositives)
		printf ("%d\n", posCnt);
	else if (countNegatives)
		printf ("%d\n", negCnt);
	}

//----------
//
// enough_overlap--
//  Determine if a feature has 'enough' overap.
//
//----------
//
// Arguments:
//	int		overlap:	The amount of overlap the feature contains.
//  int		featureLen: The length of the feature.
//
// Returns:
//  true if the overlap is enough, i.e. if it is at least as much as the
//  threshold.
//
//----------

static int enough_overlap
   (int		overlap,
	int		featureLen)
	{
	// if the threshold is simple, just compare overlap to it

	if (overlapDenom == 0)
		return (overlap >= overlapThresh);

	// otherwise, the threshold is a fraction of the feature length;  we want
	// to test
	//
	//		overlap       overlapThresh
	//		---------- >= -------------
	//		featureLen    overlapDenom
	//
	// so we mulitply by the denominators to get the equivalent test
	//
	//		overlap * overlapDenom >= overlapThresh * featureLen

	return (((long long) overlap    * overlapDenom)
	     >= ((long long) featureLen * overlapThresh));
	}

//----------
//
// print_distribution--
//	Print the positions distribution.
//
//----------
//
// Arguments:
//	vector*	posDistrib:	A vector of ints in which to accumulate the positions
//						.. distribution.
//	FILE*	distF:		The file to receive the positions distribution.
//
// Returns:
//  (nothing)
//
//----------

static void print_distribution
   (vector*	posDistrib,
	FILE*	distF)
	{
	int		pos;

	for (pos=0 ; pos<posDistrib->len ; pos++)
		fprintf (distF, "%d\t%d\n", pos, posDistrib->val[pos]);

	}

