//-------+---------+---------+---------+---------+---------+---------+--------=
//
// File:	ranges.c								Author:		Bob Harris
//
// Support code for generic C applications
//
//----------

//----------
//
// ranges--
//	Manage lists of 'ranges'.  A range is an interval of the positive number
//	line, usually representing a subsequence of a nucleotide or peptide
//	sequence.
//
//----------
//
// Implementation
// ==============
//
// Range lists are represented by the ranges_t type (note the s), and are
// implemented as a dynamically growing vector of entries.  Each entry is of
// type range_t (note the absense of an s).  Modification of a list is expected
// go through the macro ranges_INIT() and the routines ranges_append() and
// ranges_fini().  Direct, read-only, access to the ranges_t fields len and vec
// is expected.
//
// Internally, the list is allocated from the heap, and grows dynamically as
// needed.  The routine ranges_fini() should be called to de-allocate list
// elements.  The user programmer can provide custom routines to allocate and
// de-allocate;  these are carried in the ranges_t structure and must be
// plug-in replacements for realloc() and free().
//
// Two fields are provided to be freely used by the calling programmer.  Info
// can be anything the programmer wants, and infoPtr can point to any data.
// InfoPtr data will be automatically deallocated by a call to ranges_fini()
// or range_fini().  However, if it points to data containing pointers to
// other data, the programmer must deallocate this himself.  If the caller
// also deallocates the infoPtr data, he should set infoPtr to NULL.
//
// Some routines are provided (for example, parse_range()) that allow the
// caller to read ranges from a file one by one.  The caller must take care to
// properly deallocate memory allocated for each range.  range_fini() should
// be called to do so.
//
//----------
//
// Here is one example of how to make use of a range list.  In this case, the
// entire list is read into memory.
//
//		char*    filename = "whatever";
//		ranges_t myranges = ranges_INIT (ckrealloc, ckfree);
//		range_t* rng;
//		int      ix;
//
//		ranges_readfile (filename, &myranges);
//
//		for (ix=0 ; ix<r->len ; ix++)
//			{
//			rng = &r->vec[ix];
//			.. do something with rng
//			}
//
//		ranges_fini (&myranges);
//
// Here is a second example of how to make use of a range list.  Rather than
// read the entire list into memory, ranges are read and operated upon one by
// one.
//
//		FILE*   f;
//		char    buffer[.. some large size];
//		range_t rng;
//
//		f = fopen (..);
//
//		while (true)
//			{
//			if (gets ((buffer, sizeof(buffer), inFile) == NULL)
//				break;
//		    if (s[strlen(s)-1] == '\n')
//		        s[strlen(s)-1] = 0;
//
//			if (!parse_range (buffer, &rng, -1))
//				continue;
//
//			.. do something with rng
//
//			range_fini (&rng);
//			}
//
//		fclose (f);
//
//----------

#include <stdlib.h>				// standard C stuff
#define  true  1				// oughta be in stdlib.h
#define  false 0				// oughta be in stdlib.h
#include <stdio.h>				// standard C i/o stuff
#include <ctype.h>				// standard C upper/lower stuff
#include "get_line.h"			// Scott Schwartz's line reading stuff
#include "util.h"				// Scott Schwartz's handy utility stuff
#include "bobutil.h"			// Bob's handy utility stuff

#include "ranges.h"				// interface to this module

//----------
//
// ranges_fini--
//	Finish using a list of ranges.  Space for the list is de-allocated, and
//	the list length is zero'd.
//
//----------
//
// Arguments:
//	ranges_t*	r:	The range list.
//
// Returns:
//	(nothing)
//
//----------

static void ranges_fini_vec (ranges_t* r);
static void ranges_fini_vec (ranges_t* r)
	{
	if ((r->free != NULL) && (r->vec != NULL))
		{ r->free (r->vec);  r->vec = NULL;  r->max = r->len = 0; }
	}


void ranges_fini
   (ranges_t*	r)
	{
	int			ix;

	r->len = 0;

	if (r->free != NULL)
		{
		for (ix=0 ; ix<r->len ; ix++)
			range_fini (&r->vec[ix]);

		if (r->header != NULL) { r->free (r->header);  r->header = NULL; }
		ranges_fini_vec (r);
		}
	}

//----------
//
// range_fini--
//	Finish using a range.  'Extra' and 'info' space for it is de-allocated.
//
//----------
//
// Arguments:
//	range_t*	ent:	The range.
//
// Returns:
//	(nothing)
//
//----------

void range_fini
   (range_t*	ent)
	{
	if (ent->extra   != NULL) free (ent->extra);
	if (ent->infoPtr != NULL) free (ent->infoPtr);
	}

//----------
//
// ranges_need, ranges_need_exact--
//	Make sure a list of ranges has enough entries.  If it doesn't, space is
//	allocated so that it does.
//
//----------
//
// Arguments:
//	ranges_t*		r:			The range list.
//	unsigned int	howMany:	The number of entries desired.
//
// Returns:
//	true if successful;  false otherwise.
//
//----------

static int ranges_allocate (ranges_t* r, unsigned int howMany, int exactly);


int ranges_need (ranges_t* r, unsigned int howMany)
	{ return ranges_allocate (r, howMany, /*exactly*/ false); }

int ranges_need_exact  (ranges_t* r, unsigned int howMany)
	{ return ranges_allocate (r, howMany, /*exactly*/ true); }

static int ranges_allocate
   (ranges_t*		r,
	unsigned int	howMany,
	int				exactly)
	{
	range_t*		vec;
	unsigned int	exact;

	// if the list is empty, allocate enough to satisfy this request

	if (r->vec == NULL)
		{
		// on first attempt to allocate, make sure we have routines to do so

		if (r->free  == NULL) r->free  = free;
		if (r->alloc == NULL) r->alloc = realloc;

		// allocate (see note [1] below)

		exact = howMany;
		if (!exactly)
			howMany = 5 + ((4 * howMany) / 3);

		vec = (range_t*) r->alloc (NULL, howMany * sizeof(range_t));
		if (vec == NULL)
			{
			if (exactly)
				return false;
			howMany = exact;
			vec = (range_t*) r->alloc (NULL, howMany * sizeof(range_t));
			if (vec == NULL)
				return false;
			}

		r->vec = vec;
		r->len = 0;
		r->max = howMany;

		return true;
		}

	// if we already have enough allocated, just return

	if (r->max >= howMany)
		return true;

	// otherwise, we have to 'grow';  [1] we try to allocate about 33% more than
	// requested, so that we won't overwhelm the heap manager if we have
	// repeated calls requesting small growth;  if this fails, we just allocate
	// exactly as much as is needed

	exact = howMany;
	if (!exactly)
		howMany = 5 + ((4 * howMany) / 3);

    vec = (range_t*) r->alloc (r->vec, howMany * sizeof(range_t));
	if (vec == NULL)
		{
		if (exactly)
			return false;
		howMany = exact;
		vec = (range_t*) r->alloc (r->vec, howMany * sizeof(range_t));
		if (vec == NULL)
			return false;
		}

    r->vec = vec;
	r->max = howMany;

	return true;
	}

//----------
//
// ranges_more--
//	Make sure a list of ranges has more than enough entries.  If it doesn't,
//	space is allocated so that it does.
//
//----------
//
// Arguments:
//	ranges_t*		r:			The range list.
//	unsigned int	howMany:	The number of additional entries desired,
//								.. beyond the current length of the list.
//
// Returns:
//	true if successful;  false otherwise.
//
//----------

int ranges_more
   (ranges_t*		r,
	unsigned int	howMany)
	{
	return ranges_allocate (r, r->len + howMany, /*exactly*/ false);
	}

//----------
//
// ranges_append--
//	Append a range to a list of ranges.
//
//----------
//
// Arguments:
//	ranges_t*	r:			The range list.
//	range_t*	newRange:	The range to add.
//
// Returns:
//	true if successful;  false otherwise.
//
// Notes:
//  - The range is copied, but pointer fields (such as newRange->extra and
//    newRange->infoPtr) are copied as pointers (i.e. no new memory is allocated).  So
//    newRange should *not* be subjected to range_fini().
//
//----------

int ranges_append
   (ranges_t*	r,
	range_t*	newRange)
	{
	if (!ranges_more (r, 1))
		return false;

	r->vec[r->len++] = *newRange;
	return true;
	}

//----------
//
// ranges_fit--
//	Reduce the memory allocated for a list of ranges.
//
//----------
//
// Arguments:
//	ranges_t*	r:	The range list.
//
// Returns:
//	true if successful;  false otherwise (this only means that reduction
//	failed, but the list will still be valid).
//
//----------

int ranges_fit
   (ranges_t*	r)
	{
	range_t*	vec;

	vec = (range_t*) r->alloc (r->vec, r->len * sizeof(range_t));
	if (vec == NULL)
		return false;

	r->max = r->len;
	r->vec = vec;

	return true;
	}

//----------
//
// ranges_readfile--
//	Open a file, read a list of ranges from it, and close it.
//
// For the format of the file, see the routine ranges_read().
//
//----------
//
// Arguments:
//	const char*	fname:	The name of the stream to read.
//	ranges_t*	r:		The range list to store into.  This is expected to be
//						.. an empty list, such as one that has been assigned
//						.. ranges_INIT.
//
// Returns:
//	(nothing;  failures result in a program exit)
//
//----------

void ranges_readfile
   (const char*	fname,
	ranges_t*	r)
	{
	FILE*		f;

	f = ckopen (fname, "rt");
	ranges_read (f, r);
	fclose (f);
	}

//----------
//
// ranges_read--
//	Read a list of ranges from a file.
//
// Range lists are expected to appear in the file in this format (more or
// less):
//
//		# TATA Box
//		chr1	26363519	26364118
//		chr1	45462486	45463085
//		chr1	156773859	156774458
//		chr3	150512857	150513456
//		chr4	41173921	41174520
//		 ...
//
// Any lines beginning with '#' are treated as comments.  If the first line in
// the file is a comment, it is saved as the list's header.
//
// See the notes in the routine parse_range() for more details.
//
//----------
//
// Arguments:
//	FILE*		f:	The stream to read;  it is assumed to have been opened
//					.. as text, and positioned at the beginning.
//	ranges_t*	r:	The range list to store into.  This is expected to be
//					.. an empty list, such as one that has been assigned 
//					.. ranges_INIT.
//
// Returns:
//	(nothing;  failures result in a program exit)
//
//----------

void ranges_read
   (FILE*			f,
	ranges_t*		r)
	{
	char*			buffer    = NULL;
	unsigned long	bufferLen = 0;
	unsigned long	lineLen   = 0;
	int				lineNumber;
	char*			s;
	range_t			entry;

	lineNumber = 0;
	while (true)
		{
		// read the next line;  note that get_line does not allow us to
		// distinguish an error from an end-of-file

		lineLen = get_line (&buffer, &bufferLen, f);
		if (lineLen == (unsigned long) -1) break;
		lineNumber++;
		if (lineLen == 0) continue;
		if ((buffer[lineLen-1] == '\n')
		 || (buffer[lineLen-1] == '\r'))
			buffer[--lineLen] = 0;
		if (lineLen == 0) continue;

		// if it's a comment, ignore it;  but if it's the first line, then use
		// it as the header

		if (buffer[0] == '#')
			{
			if (lineNumber == 1)
				{
				s = skip_whitespace (buffer+1);
				if (*s != 0)
					r->header = copy_string (s);
				}
			continue;
			}

		// parse the range and add it to the list

		if (!parse_range (buffer, &entry, lineNumber))
			continue;	// ignore things like comments

		if (!ranges_append (r, &entry))
			fatalf ("*** ERROR: Allocation failure reading range list file (line %d). ***\n\n",
			        lineNumber);
		}

	// success

	if (r->len != 0)
		ranges_fit (r);

	ckfree (buffer);
	}

//----------
//
// parse_range--
//	Parse a line of text that represents a range.
//
// Each line is of the format
//  <name> <start> <end> [<dir>] [<extra>]
//
// Fields can be separated with whitespace and/or a comma.  Non-numeric fields
// are allowed to have quotes (which are skipped).
//
// The name is allowed to be anything, but if it begins with "chr", it is
// encoded into the range's chr field.
//
// Start and end are numbers.  Start can be after end, in which case this
// routine flips them and sets the strand flag in the entry.
//
// Dir can be + or - and, if present, is copied into the strand flag.  Note
// that there are two ways to set the strand flag;  if both are used they
// cancel each other.  If dir contains additional characters (such as strand
// direction for other species), they are ignored.
//
// Generally, anything after the dir field (or after the end field if dir is
// not present) is copied as extra.  The caller should call range_fini() to
// de-allocate such storage.  In some cases not all of what's left over is
// copied.  If the line has a format similar to this:
//	'chr9',15814,15821,'-++',0.87,'F$STRE_01'
// the the 0.87 is skipped, and only F$STRE_01 is copied as extra.
//
//----------
//
// Arguments:
//	char*		s:			The line to parse.  Note that this should *not*
//							.. contain a newline character.
//	range_t*	ent:		The range to store into.
//	int			lineNumber:	The number of the line being parsed.  This is used
//							.. solely for reporting errors.  If tha caller
//							.. doesn't have a line number, -1 should be used.
//
// Returns:
//	true if we have an acceptable range;  false if the line contained a
//	comment.  Note that errant lines cause the *program* to exit.
//
//----------

static char* skip_white_comma_ok (char* s);
static char* skip_white_comma_ok (char* s)
	{
	s = skip_whitespace (s);
	if (*s == ',') s = skip_whitespace (s+1);
	return s;
	}

char* skip_black_plus_comma_or_colon (char* s);
char* skip_black_plus_comma_or_colon (char* s)
	{
	for ( ; (*s!=0)&&(*s!=',')&&(*s!=':')&&(!isspace(*s)) ; s++) ;
	if      (*s == ',') s = skip_whitespace (s+1);
	else if (*s == ':') s++;
	return s;
	}

char* skip_quote (char* s);
char* skip_quote (char* s)
	{
	if ((*s == '\'') || (*s == '"')) s++;
	return s;
	}

char* skip_signs (char* s);
char* skip_signs (char* s)
	{
	for ( ; (*s!=0)&&((*s=='+')||(*s=='-')) ; s++) ;
	return s;
	}

char* skip_real_number (char* s);
char* skip_real_number (char* s)
	{
	if ((*s == '-') || (*s == '+')) s++;
	for ( ; (*s!=0)&&(isdigit(*s)) ; s++) ;
	if (*s == '.')
		{
		s++;
		for ( ; (*s!=0)&&(isdigit(*s)) ; s++) ;
		}
	return s;
	}

int parse_range
   (char*			s,
	range_t*		ent,
	int				lineNumber)
	{
	int				numItems, itemsLen;
	char*			s2;
	unsigned long	t;
	int				len;

	// if it's a blank line or comment, ignore it

	if ((*s == 0) || (*s == '#'))
		return false;

	// skip to where the numbers should be

	s  = skip_whitespace (s);
	s2 = skip_black_plus_comma_or_colon (s);

	// parse the numbers

	numItems = sscanf (s2, " %lu-%lu%n",  // chrom:start-end
	                       &ent->start, &ent->end, &itemsLen);

	if (numItems != 2) // chrom start end (must come after chrom:start-end
	                   // because scanf gets confused otherwise
		numItems = sscanf (s2, " %lu %lu%n",
		                       &ent->start, &ent->end, &itemsLen);

	if (numItems != 2)  // chrom,start,end
		numItems = sscanf (s2, " %lu,%lu%n",
		                       &ent->start, &ent->end, &itemsLen);

	if (numItems != 2)
		{
		if (lineNumber == -1)
			fatalf ("*** ERROR: Invalid range list file format.  \"%s\" ***\n\n",
			        s);
		else
			fatalf ("*** ERROR: Invalid range list file format (line %d).  \"%s\" ***\n\n",
			        lineNumber, s);
		}

	s2 += itemsLen;
	s2 = skip_white_comma_ok (s2);

	// parse the chromosome

	s = skip_quote (s);
	ent->chr = str_to_chromosome (s);

	// parse the strand;  note that if a + or - is followed by something other
	// that whitespace or more + or -, it is *not* a strand

	ent->strand = 0;	// (no direction)
	ent->extra  = NULL;

	s2 = skip_quote (s2);
	if ((*s2 == '+') || (*s2 == '-'))
		{
		if ((isspace (s2[1]))
		 || (s2[1] == 0) || (s2[1] == '+') || (s2[1] == '-'))
			{
			ent->strand = *(s2++);
			s2 = skip_signs (s2);
			}
		}

	s2 = skip_quote (s2);
	s2 = skip_white_comma_ok (s2);

	if (ent->start > ent->end)
		{
		t          = ent->start;
		ent->start = ent->end;
		ent->end   = t;
		if (ent->strand == '-') ent->strand = '+';
		                   else ent->strand = '-';
		}

	// if the first extra field is just a number, skip it to

	// s3 = skip_quote (s2);
	// s4 = skip_real_number (s3);
	// if (s4 != s3)
	//    {
	//    s2 = skip_quote (s4);
	//    s2 = skip_white_comma_ok (s2);
	//    }

	// anything extra, just copy it

	s2 = skip_quote (s2);

	if (*s2 != 0)
		{
		len = strlen (s2);
		while ((len > 0) && ((s2[len-1] == '\'')
		                  || (s2[len-1] == '"')
		                  || (s2[len-1] == ',')))
			{ len--; }
		if (len > 0)
			ent->extra = copy_substring (s2, len);
		}

	// clear the user-defined fields

	ent->info    = 0;
	ent->infoPtr = NULL;

	return true;
	}

//----------
//
// ranges_fprintf--
//	Print a list of ranges.
// ranges_fprintf_info--
//	Print a list of ranges, including the info field.
//
//----------
//
// Arguments:
//	FILE*		f:		The stream to print to.
//	ranges_t*	r:		The range list to print.
//	char*		format:	The printf format.  This should contain the following
//						.. specifiers, in the order below.  Normally the
//						.. format string should also finish with a newline.
//						    %s    chr
//							%lu   start
//							%lu   end
//							%s    reverse
//							%ld   info (only for xxx_info calls)
//							%s    extra
//
// Returns:
//	(nothing)
//
//----------

void ranges_fprintf
   (FILE*		f,
	ranges_t*	r,
	char*		format)
	{
	int			ix;

	if ((r->vec == NULL)
	 || (r->len == 0))
		{
		fprintf (f, "(empty range list)\n");
		return;
		}

	for (ix=0 ; ix<r->len ; ix++)
		range_fprintf (f, &r->vec[ix], format);
	}


void ranges_fprintf_info
   (FILE*		f,
	ranges_t*	r,
	char*		format)
	{
	int			ix;

	if ((r->vec == NULL)
	 || (r->len == 0))
		{
		fprintf (f, "(empty range list)\n");
		return;
		}

	for (ix=0 ; ix<r->len ; ix++)
		range_fprintf_info (f, &r->vec[ix], format);
	}

//----------
//
// range_fprintf--
//	Print a range.
// ranges_fprintf_info--
//	Print a range, including the info field.
//
//----------
//
// Arguments:
//	FILE*		f:		The stream to print to.
//	range_t*	ent:	The range to print.
//	char*		format:	The printf format;  see ranges_fprintf() for more
//						.. details.
//
// Returns:
//	(nothing)
//
//----------

void range_fprintf
   (FILE*		f,
	range_t*	ent,
	char*		format)
	{
	char*		strandStr;

	switch (ent->strand)
		{
		case '+': strandStr = "+";   break;
		case '-': strandStr = "-";   break;
		default:
		case 0:   strandStr = "";    break;
		}

	fprintf (f, format,
	         chromosome_to_str (ent->chr),
	         ent->start, ent->end, strandStr,
	         (ent->extra != NULL)? ent->extra : "");
	}

void range_fprintf_info
   (FILE*		f,
	range_t*	ent,
	char*		format)
	{
	char*		strandStr;

	switch (ent->strand)
		{
		case '+': strandStr = "+";   break;
		case '-': strandStr = "-";   break;
		default:
		case 0:   strandStr = "";    break;
		}

	fprintf (f, format,
	         chromosome_to_str (ent->chr),
	         ent->start, ent->end, strandStr, ent->info,
	         (ent->extra != NULL)? ent->extra : "");
	}

//----------
//
// ranges_sort--
//	Sort a list of ranges.
//
//----------
//
// Arguments:
//	ranges_t*	r:		The range list to sort.
//	int (*compar)(const void*, const void*): A comparison function,
//						.. suitable for use by qsort.  If this is NULL,
//						.. ranges_ascending is used.
//
// Returns:
//	(nothing)
//
//----------

void ranges_sort
   (ranges_t*	r,
	int (*compar)(const void*, const void*))
	{
	if (compar == NULL)
		qsort (r->vec, r->len, sizeof(range_t), ranges_ascending);
	else
		qsort (r->vec, r->len, sizeof(range_t), compar);
	}

//----------
// [[-- comparison function for the standard c function qsort --]]
//
// ranges_ascending--
//	Compare two ranges, to order the chromosomes and start indexes from low to
//	high.
//
//----------
//
// Arguments:
//	const void* n1:	One range (really const range_t*).
//	const void* n2:	Another   (really const range_t*).
//
// returns:
//	> 0 => n1 is greater than n2.
//	= 0 => n1 and n2 are the same.
//	< 0 => n1 is less than n2.
//
//----------

int ranges_ascending
   (const void*	n1,
	const void*	n2)
	{
	range_t*	r1 = (range_t*) n1;
	range_t*	r2 = (range_t*) n2;
	char*		extra1;
	char*		extra2;

	// compare chomosomes first

	if      (r1->chr > r2->chr) return  1;
	else if (r1->chr < r2->chr) return -1;

	// break ties with start indexes

	if      (r1->start > r2->start) return  1;
	else if (r1->start < r2->start) return -1;

	// ... then with end indexes

	if      (r1->end > r2->end) return  1;
	else if (r1->end < r2->end) return -1;

	// ... then with direction flag (forward before reverse)

	if      (r1->strand > r2->strand) return  1;
	else if (r1->strand < r2->strand) return -1;

	// ... then with programmer's info

	if      (r1->info > r2->info) return  1;
	else if (r1->info < r2->info) return -1;

	// ... then with any 'extra info'

	extra1 = (r1->extra != NULL)? r1->extra : "";
	extra2 = (r2->extra != NULL)? r2->extra : "";

	return strcmp (extra1, extra2);
	}

//----------
//
// ranges_exclude--
//	Exclude any portion of a list of ranges that intersects with another list.
//	Any range or subrange that intersects with the to-be-excluded list is
//	removed.
//
//----------
//
// Arguments:
//	ranges_t*	r:			The range list to modify.  This is expected to
//							.. be in increasing order by chromosomes and
//							.. start indexes.  Upon exit this will contain the
//							.. subranges of the original which do not intersect
//							.. the the other range list.  Note that if the
//							.. input ranges contain overlaps, so may the output.
//	ranges_t*	exclude:	The range list to exclude.  This is expected to
//
// Returns:
//	(nothing)
//
//----------

void ranges_exclude
   (ranges_t*		r,
	ranges_t*		exclude)
	{
	ranges_t		result = ranges_INIT (ckrealloc, ckfree);
	range_t			rng;
	int				rIx, xIx, markIx;
	unsigned long	end;

	// look at each range in the list, removing any overlapping intervals,
	// and creating a new list for what's left

	xIx = 0;
	for (rIx=0 ; rIx<r->len ; rIx++)
		{
		rng = r->vec[rIx];

		// advance exclude-interval until it no long precedes the interval of
		// interest

		while ((xIx < exclude->len) && (exclude->vec[xIx].chr < rng.chr))
			xIx++;
		while ((xIx < exclude->len) && (exclude->vec[xIx].end < rng.start))
			xIx++;

		// remove any exclude-intervals from the interval of interest

		markIx = xIx;

		while ((xIx < exclude->len)
		    && (exclude->vec[xIx].chr == rng.chr)
		    && (exclude->vec[xIx].start <= rng.end))
			{
			if (exclude->vec[xIx].start > rng.start)
				{
				end     = rng.end;
				rng.end = exclude->vec[xIx].start-1;
				if (!ranges_append (&result, &rng))
					fatalf ("*** ERROR: Allocation failure excluding ranges. ***\n\n");
				rng.end = end;
				}

			if (exclude->vec[xIx].end >= rng.start)
				rng.start = exclude->vec[xIx].end+1;

			if (rng.start > rng.end) break; // (interval of interest now empty)
			xIx++;
			}

		if (rng.start <= rng.end)
			{
			if (!ranges_append (&result, &rng))
				fatalf ("*** ERROR: Allocation failure excluding ranges. ***\n\n");
			}

		xIx = markIx;
		}

	// copy the new list into the old one, and dispose of the old one

	ranges_fini_vec (r);

	r->vec = result.vec;
	r->len = result.len;
	r->max = result.max;
	if (r->len != 0) ranges_fit (r);
	}

//----------
//
// ranges_in_interval--
//	Determine if a list of ranges intersects with an interval.
//
//----------
//
// Arguments:
//	ranges_t*		r:			The range list.  This is expected to be in
//								.. be in increasing order by chromosomes and
//								.. start indexes.
//	char			chr:		The chromosome the interval is in.
//	unsigned long	firstIndex:	The first index in the interval.
//	unsigned long	lastIndex:	The final index in the interval (inclusive).
//								.. Note that lastIndex >= firstIndex.
//	char			strand:		sequence direction
//								.. '+' => forward strand
//								.. '-' => reverse strand
//								.. 0   => both strands
//
// Returns:
//	The number of entries in the range list that
//	  (a) are in the same direction as the interval.
//  and
//	  (b) intersect, even partially, with the interval.
//
//----------

int range_in_ranges
   (range_t*		rng,
	ranges_t*		r)
	{
	return ranges_in_interval (r, rng->chr, rng->start, rng->end, rng->strand);
	}

int ranges_in_interval
   (ranges_t*		r,
	char			chr,
	unsigned long	firstIndex,
	unsigned long	lastIndex,
	char			strand)
	{
	int				hits;
	int				rIx;
	range_t*		ent;

	// make sure we have ranges in the list

	if (r->len < 1)
		return 0;

	// do linear search (binary search would be faster)

	hits = 0;

	for (rIx=0 ; rIx<r->len ; rIx++)
		{
		ent = &r->vec[rIx];

		if (ent->chr >= chr)
			break;
		}

	for ( ; rIx<r->len ; rIx++)
		{
		ent = &r->vec[rIx];

		if ((ent->chr   > chr)
		 || (ent->start > lastIndex))
			break;

		if (((ent->start <  firstIndex) && (ent->end >= firstIndex))
		 ||  (ent->start >= firstIndex))
			{
			if ((ent->strand == 0)
			 || (strand == 0)
			 || (ent->strand == strand))
				hits++;
			}
		}

	return hits;
	}

//----------
//
// first_range_in_interval--
//	Find the 'first' intersection of an interval with a list of ranges.
//
//----------
//
// Arguments:
//	ranges_t*		r:			The range list.  This is expected to be in
//								.. be in increasing order by chromosomes and
//								.. start indexes.
//  int				rIx:		The index (into r->vec[]) of the first range
//								.. to consider.  Ranges are considered by
//								.. increasing index until an intersection is
//								.. encountered.
//	char			chr:		The chromosome the interval is in.
//	unsigned long	firstIndex:	The first index in the interval.
//	unsigned long	lastIndex:	The final index in the interval (inclusive).
//								.. Note that lastIndex >= firstIndex.
//	char			strand:		sequence direction
//								.. '+' => forward strand
//								.. '-' => reverse strand
//								.. 0   => both strands
//
// Returns:
//  The index (into r->vec[]) of the first range that intersects with the
//  interval;  -1 if no range intersects.
//
//----------

int first_range_in_interval
   (ranges_t*		r,
	int				rIx,
	char			chr,
	unsigned long	firstIndex,
	unsigned long	lastIndex,
	char			strand)
	{
	range_t*		ent;

	// continue our linear search

	for ( ; rIx<r->len ; rIx++)
		{
		ent = &r->vec[rIx];

		if (ent->chr >= chr)
			break;
		}

	for ( ; rIx<r->len ; rIx++)
		{
		ent = &r->vec[rIx];

		if ((ent->chr   > chr)
		 || (ent->start > lastIndex))
			break;

		if (((ent->start <  firstIndex) && (ent->end >= firstIndex))
		 ||  (ent->start >= firstIndex))
			{
			if ((ent->strand == 0)
			 || (strand == 0)
			 || (ent->strand == strand))
				return rIx;
			}
		}

	return -1;
	}

//----------
//
// ranges_overlap_interval--
//  Determine the total amount of overlap between an interval and a list of
//  ranges.  Overlap is measured as a count of base pairs.
//
// NOTE: It is assumed that the list of ranges contains no overlap with itself.
//
//----------
//
// Arguments:
//	ranges_t*		r:			The range list.  This is expected to be in
//								.. be in increasing order by chromosomes and
//								.. start indexes.  It is also assmued to
//								.. contain no self-overlaps.
//	char			chr:		The chromosome the interval is in.
//	unsigned long	firstIndex:	The first index in the interval.
//	unsigned long	lastIndex:	The final index in the interval (inclusive).
//								.. Note that lastIndex >= firstIndex.
//	char			strand:		sequence direction
//								.. '+' => forward strand
//								.. '-' => reverse strand
//								.. 0   => both strands
//
// Returns:
//  The number of base pairs in the list of ranges that overlap the interval,
//  provided that the range is in the same direction as the interval.
//
//----------

int ranges_overlap_range
   (range_t*		rng,
	ranges_t*		r)
	{
	return ranges_overlap_interval (r, rng->chr, rng->start, rng->end, rng->strand);
	}

int ranges_overlap_interval
   (ranges_t*		r,
	char			chr,
	unsigned long	firstIndex,
	unsigned long	lastIndex,
	char			strand)
	{
	int				overlapStart, overlapEnd, totalOverlap;
	int				rIx;
	range_t*		ent;

	// make sure we have ranges in the list

	if (r->len < 1)
		return 0;

	// locate the right chromosome

	totalOverlap = 0;

	for (rIx=0 ; rIx<r->len ; rIx++)
		{
		ent = &r->vec[rIx];

		if (ent->chr >= chr)
			break;
		}

	// scan the chromosome tallying overlap

	for ( ; rIx<r->len ; rIx++)
		{
		ent = &r->vec[rIx];

		if ((ent->chr   > chr)
		 || (ent->start > lastIndex))
			break;		// (we've passed the interval)

		if (ent->end < firstIndex)
			continue;   // (doesn't overlap the interval)

		if ((ent->strand != 0)
		 && (strand != 0)
		 && (ent->strand != strand))
			continue;   // (direction not 'same' as the interval's)

		overlapStart = ent->start;
		overlapEnd   = ent->end;

		if (ent->start < firstIndex) overlapStart = firstIndex;
		if (ent->end   > lastIndex)  overlapEnd   = lastIndex;

		totalOverlap += overlapEnd+1 - overlapStart;
		}

	return totalOverlap;
	}

